function mininvasion
% solve "minimal invasion problem" using semi-smooth Newton method
%
% Christian Clason (christian.clason@uni-graz.at)
% June 10, 2010

%% setup
N     = 64;         % grid size
alpha = 1e-6;       % control penalty (L2)
M     = 4;          % number of control domains
maxit = 200;        % max iterations in semi-smooth Newton method
tol   = 1e-9;       % terminate if feasibility gap below tol

% constraints: -beta2 c + psi2 <= y <= beta1 c + psi1
beta1 = 1;
beta2 = 0;
psi1  = 0;
psi2  = 0;

upper = 1; % upper constraint active
lower = 1; % lower constraint active

% coefficients of differential operator
nu = 0.1;           % diffusion
b  = -[0,1];        % convection

% setup grid on [-1,1]^2
[x,y] = meshgrid(linspace(-1,1,N));
h     = 2/N;

% setup observation and control subdomains
om0 = (((x).^2+y.^2)<1/64)+0*(x<-.75);   % city (observation domain)

om1 = (((x+.5).^2+(y-.5).^2)<1/32);      % well ll (control domain)
om2 = (((x+.5).^2+(y+.5).^2)<1/32);      % well ul (control domain)
om3 = (((x-.5).^2+(y+.5).^2)<1/32);      % well ur (control domain)
om4 = (((x-.5).^2+(y-.5).^2)<1/32);      % well lr (control domain)
om  = [om1(:)';om2(:)';om3(:)';om4(:)']';

dplot = @(n,f,s) domainplot_(n,f,s,N,x,y,om0,om1|om2|om3|om4);

% setup operators
Dxx = gallery('poisson',N)*1/h^2; % -Laplace
e   = ones(N,1);
Dx  = 1/h*spdiags([-e e],-1:0,N,N);
Ix  = speye(N,N);

A   = nu*Dxx+b(1)*kron(Ix,Dx)+b(2)*kron(Dx,Ix);

% setup forcing term (pollution)
f = 10*(1+y).*(x>.75); f = f(:);

% uncontrolled state
y0 = A\f;

% plot geometric setup, uncontrolled state
display(sprintf('|y0|_om0 \t= %g',max(abs(y0(:)).*om0(:))));
dplot(1,f,'pollution and control, observation domains');
text(.12,-.12,'\omega_0','FontSize',12);text(.7,-.7,'\omega_3','FontSize',12);text(.7,.7,'\omega_4','FontSize',12);
text(-.75,.75,'\omega_1','FontSize',12);text(-.8,-.8,'\omega_2','FontSize',12);
dplot(2,y0,'uncontrolled state y^0');

% precompute Hessian (iteration-independent components)
Luu = alpha*speye(M,M);
Luc = spalloc(M,1,0);
Luy = spalloc(M,N*N,0);
Lup = - om';
Lcp = spalloc(1,N*N,0);
Lyp = A';
Lpp = spalloc(N*N,N*N,0);

%% Solve optimality system

% initialize iterates
u = zeros(M,1);
c = 0;
y = zeros(N*N,1);
p = zeros(N*N,1);

% continuation in Moreau-Yosida-penalty gamma
for g = 10.^(0:16)
    fprintf('Compute minimizer for gamma=%1.0e\n',g);
    
    % semi-smooth Newton method
    iter   = 0;
    update = 1e99;
    
    while (update) && (iter < maxit)
        
        % compute active sets
        l1 = y-(beta1*c+psi1);      % Lagrange multiplier \lambda_1
        l2 = y-(-beta2*c+psi2);     % Lagrange multiplier \lambda_2
        ap = (l1>0)*upper;          % active set for upper constraint
        am = (l2<0)*lower;          % active set for lower constraint
        
        % compute Hessian (iteration-dependent components)
        Lcc = 1 + g*(om0(:)'*(beta1^2*ap+beta2^2*am));
        Lcy = g*(om0(:).*(-beta1*ap+beta2*am))';
        Lyy = spdiags(g*(ap+am).*om0(:),0,N*N,N*N);
        H = [Luu  Luc  Luy  Lup;
            Luc' Lcc  Lcy   Lcp;
            Luy' Lcy' Lyy   Lyp;
            Lup' Lcp' Lyp'  Lpp];
        % compute gradient
        Lu = alpha*u - om'*p;
        Lc = c + g*om0(:)'*(-beta1*ap.*l1+beta2*am.*l2);
        Ly = g*(ap.*l1+am.*l2).*om0(:) + A'*p;
        Lp = A*y - f - om*u;
        rhs = -[Lu;Lc;Ly;Lp];
        
        % Newton step
        dx = H\rhs;
        
        u = u+dx(1:M);
        c = c+dx(M+1:M+1);
        y = y+dx(M+2:N^2+M+1);
        p = p+dx(M+2+N^2:end);
        
        % terminate if active sets coincide
        if iter == 0
            update = 1e99;
        else
            update = nnz((ap-ap_old) | (am-am_old));
            fprintf('It# %i: update = %i\n',iter,update);
        end
        ap_old = ap;
        am_old = am;
        
        iter = iter+1;
        
    end
    
    %% output
    fprintf('u^* \t= %g\n',u);
    fprintf('c^* \t= %g\n',c);
    fprintf('y_max  = %g\n',max(y(om0>0)));
    fprintf('y_min  = %g\n',min(y(om0>0)));
    
    dplot(3,y,'optimal state y^*')
    dplot(4,y-y0,'difference y^*-y^0')
    dplot(5,p,'optimal Lagrange multiplier p^*')
    dplot(6,om*u,'optimal control u^*')

	% terminate continuation if feasibility within tolerance
    if ((max((y(:)-beta1*c-psi1).*om0(:))<tol) && (min((y(:)+beta2*c-psi2).*om0(:)))<tol)
        break
    end
    
end

end % main function

function domainplot_(n,f,s,N,x,y,om0,oms)
figure(n);
surf(x,y,reshape(f,N,N)); set(gca,'FontSize',12);
shading interp; camlight headlight; lighting gouraud; alpha(0.67);
hold on;
contourz(x,y,reshape(f,N,N),oms,[1,1],'k');
contourz(x,y,reshape(f,N,N),om0,[1,1],'r');
hold off
title(s);
xlabel('x');ylabel('y');
drawnow;
end % domainplot_ function